from ekp_sdk.util.clean_null_terms import clean_null_terms


def Tabs(
    children,
    class_name=None,
    style=None,
    tooltip=None,
):
    return {
        "_type": "Tabs",
        "props": clean_null_terms({
            "children": children,
            "className": class_name,
            "style": style,
            "tooltip": tooltip,
        })
    }


def Tab(
    label,
    children,
):
    return {
        "label": label,
        "children": children
    }



from ekp_sdk.db.mg_client import MgClient
from pymongo import DESCENDING, UpdateOne
import time


class ContractTransactionsRepo:
    def __init__(
            self,
            mg_client: MgClient
    ):
        self.mg_client = mg_client

        self.collection = self.mg_client.db['contract_transactions']
        self.collection.create_index("hash", unique=True)
        self.collection.create_index([("blockNumber", DESCENDING)])
        self.collection.create_index([("timeStamp", DESCENDING)])
        self.collection.create_index([("methodId", DESCENDING)])
        self.collection.create_index("source_contract_address")
        self.collection.create_index("chain")

    def get_latest(self, contract_address):
        return list(
            self.collection.find(
                {"source_contract_address": contract_address}
            )
                .sort("blockNumber", -1).limit(1)
        )

    def find_since_block_number(
        self, 
        block_number, 
        limit, 
        method_id=None,
        source_contract_address=None        
        ):
        start = time.perf_counter()
        query = {"blockNumber": {"$gte": block_number}}
        
        if method_id is not None:
            query["methodId"] = method_id
            
        if source_contract_address is not None:
            query["source_contract_address"] = source_contract_address
            
        results = list(
            self.collection
                .find(query)
                .sort("blockNumber")
                .limit(limit)
        )

        print(
            f"⏱  [ContractTransactionsRepo.find_since_block_number({block_number})] {time.perf_counter() - start:0.3f}s")

        return results

    def save(self, trans):
        start = time.perf_counter()

        self.collection.bulk_write(
            list(map(lambda tran: UpdateOne(
                {"hash": tran["hash"]}, {"$set": tran}, True), trans))
        )

        print(f"⏱  [ContractTransactionsRepo.save({len(trans)})] {time.perf_counter() - start:0.3f}s")

    def save_logs(self, logs):
        start = time.perf_counter()

        def format_write(log):
            log_index = log["logIndex"]
            return UpdateOne(
                {"hash": log["transactionHash"]},
                {"$set": {f"logs.{log_index}": log}},
                True
            )

        self.collection.bulk_write(
            list(map(lambda log: format_write(log), logs))
        )

        print(f"⏱  [ContractTransactionsRepo.saveLogs({len(logs)})] {time.perf_counter() - start:0.3f}s")


from ekp_sdk.ui.components.chart import *
from ekp_sdk.ui.components.link import *
from ekp_sdk.ui.components.others import *
from ekp_sdk.ui.components.tabs import *
from ekp_sdk.ui.components.form import *
from ekp_sdk.ui.components.input import *



from ast import literal_eval

from ekp_sdk.db.contract_logs_repo import ContractLogsRepo
from ekp_sdk.db.contract_transactions_repo import ContractTransactionsRepo
from ekp_sdk.services import EtherscanService


class TransactionSyncService:
    def __init__(
            self,
            contract_transactions_repo: ContractTransactionsRepo,
            contract_logs_repo: ContractLogsRepo,
            etherscan_service: EtherscanService,
    ):
        self.contract_transactions_repo = contract_transactions_repo
        self.contract_logs_repo = contract_logs_repo
        self.etherscan_service = etherscan_service
        self.page_size = 1000

    async def sync_transactions(self, contract_address, start_block_number=0, etherscan_service=None, chain=None):

        if not etherscan_service:
            etherscan_service = self.etherscan_service

        latest_block_number = await etherscan_service.get_latest_block_number()

        latest_transaction = self.contract_transactions_repo.get_latest(
            contract_address
        )

        if latest_transaction is not None and len(latest_transaction):
            repo_latest_block_number = latest_transaction[0]["blockNumber"]

            if repo_latest_block_number < start_block_number:
                start_block_number = repo_latest_block_number

            start_block_number = repo_latest_block_number

        original_start_block = start_block_number

        while True:
            trans = await etherscan_service.get_transactions(contract_address, start_block_number, self.page_size)

            if len(trans) == 0:
                break

            models = []

            for tran in trans:
                block_number = int(tran["blockNumber"])

                if block_number > start_block_number:
                    start_block_number = block_number

                tran["blockNumber"] = block_number
                tran["source_contract_address"] = contract_address
                tran["confirmations"] = int(tran["confirmations"])
                tran["cumulativeGasUsed"] = int(tran["cumulativeGasUsed"])
                tran["gas"] = int(tran["gas"])
                tran["gasUsed"] = int(tran["gasUsed"])
                tran["gasPrice"] = int(tran["gasPrice"])
                tran["isError"] = tran["isError"] == "1"
                tran["timeStamp"] = int(tran["timeStamp"])
                tran["chain"] = chain
                tran["transactionIndex"] = int(tran["transactionIndex"])
                if len(tran["input"]) >= 10:
                    tran["methodId"] = tran["input"][0:10]

                models.append(tran)

            self.contract_transactions_repo.save(models)

            pc_complete = round(
                (block_number - original_start_block) * 100 /
                (latest_block_number - original_start_block),
                1
            )

            print(
                f"🐛 [Transaction Sync] - [{contract_address}] - {block_number} / {latest_block_number} ({pc_complete} %)"
            )

            if (len(trans) < self.page_size):
                break

    async def sync_logs(
        self,
        log_address,
        start_block_number=0,
        etherscan_service=None,
        chain=None,
        topic0=None
    ):

        if not etherscan_service:
            etherscan_service = self.etherscan_service

        latest_block_number = await etherscan_service.get_latest_block_number()

        latest_log = self.contract_logs_repo.get_latest(
            log_address
        )

        if latest_log is not None and len(latest_log):
            repo_latest_block_number = latest_log[0]["blockNumber"]
            if repo_latest_block_number < start_block_number:
                start_block_number = repo_latest_block_number

        original_start_block = start_block_number

        while True:
            logs = await etherscan_service.get_logs(log_address, start_block_number, topic0=topic0)

            if len(logs) == 0:
                break

            models = []

            for log in logs:
                block_number = literal_eval(log["blockNumber"])

                if block_number > start_block_number:
                    start_block_number = block_number

                log["blockNumber"] = block_number
                log["gasUsed"] = 0 if log["gasUsed"] == "0x" else literal_eval(log["gasUsed"])
                log["gasPrice"] = 0 if log["gasPrice"] == "0x" else literal_eval(log["gasPrice"])
                log["timeStamp"] = literal_eval(log["timeStamp"])

                if (log["logIndex"] == "0x"):
                    log["logIndex"] = 0
                else:
                    log["logIndex"] = literal_eval(log["logIndex"])

                if (log["transactionIndex"] == "0x"):
                    log["transactionIndex"] = 0
                else:
                    log["transactionIndex"] = literal_eval(
                        log["transactionIndex"])
                log["chain"] = chain
                models.append(log)

            self.contract_logs_repo.save(models)
            self.contract_transactions_repo.save_logs(models)

            pc_complete = round(
                (block_number - original_start_block) * 100 /
                (latest_block_number - original_start_block),
                1
            )

            print(
                f"🐛 [Log Sync] - {log_address} - {block_number} / {latest_block_number} ({pc_complete} %)"
            )

            if (len(logs) < 1000):
                break



from decouple import config
from pymongo import MongoClient

MONGO_URI = config('MONGO_URI', default='mongodb://localhost:27017/')
MONGO_DB_NAME = config('MONGO_DB_NAME', default='dalarnia')


class MgClient:
    def __init__(self, uri, db_name):
        self.client = MongoClient(uri)
        self.db = self.client[db_name]
        print("👍 Connected to MongoDb")
